/////////////////////////////////////////////////////////////////////////
// reflexive.go
/////////////////////////////////////////////////////////////////////////
// Author: Alexander M Kasprzyk
// Date: 1 April 2013
/////////////////////////////////////////////////////////////////////////
/*
Classifies the terminal Gorenstein weighted projective spaces in low dimensions.
Note that the output will *not* include PP^n.
*/

package main

import (
	"bufio"
	"flag"
	"fmt"
	"io"
	"math/big"
	"os"
	"runtime"
	"strings"
)

// The command-line arguments structure
type cmdArgs struct {
	n   int    // The dimension to search
	out string // The output file (if any)
}

// A block of memory used to reduce memory assignment
type scratchBuf struct {
	b []int64 // The chunk of memory
	n int     // The block size
	l int     // The number of remaining blocks
}

// The destination writer
var out_w io.Writer

// The size of the results channel buffer
const CHAN_BUF_SIZE = 16 * 1024

// The size of the scratch buffer
const SCRATCH_SIZE = 64 * 1024

// The size of the output writer buffer
const WRITER_BUF_SIZE = 16 * 1024

/////////////////////////////////////////////////////////////////////////
// Maths functions
/////////////////////////////////////////////////////////////////////////

// Gcd returns the GCD of the two integers a and b.
func Gcd(a int64, b int64) int64 {
	if a < 0 {
		a = -a
	}
	if a == 0 {
		return b
	}
	return Gcd(b%a, a)
}

// Lcm returns the LCM of the two integers a and b.
func Lcm(a int64, b int64) int64 {
	return (a * b) / Gcd(a, b)
}

// LcmSlice returns the LCM of the given slice of integers.
func LcmSlice(s []int64) int64 {
	b := int64(1)
	for _, a := range s {
		b = (a * b) / Gcd(a, b)
	}
	return b
}

// Ceiling returns the smallest integer bigger than or equal to a / b.
func Ceiling(a int64, b int64) int64 {
	if b < 0 {
		a = -a
		b = -b
	}
	c := (b - (a % b)) % b
	return (a + c) / b
}

// Floor returns the largest integer smaller than or equal to a / b.
func Floor(a int64, b int64) int64 {
	if b < 0 {
		a = -a
		b = -b
	}
	if a < 0 {
		c := (b + (a % b)) % b
		return (a - c) / b
	}
	return (a - (a % b)) / b
}

/////////////////////////////////////////////////////////////////////////
// Local functions
/////////////////////////////////////////////////////////////////////////

// new_scratch returns a new scratch buffer using block size n + 1.
func new_scratch(n int) *scratchBuf {
	scratch := new(scratchBuf)
	scratch.n = n
	scratch.l = SCRATCH_SIZE
	scratch.b = make([]int64, (n+1)*SCRATCH_SIZE)
	return scratch
}

// extract_from_scratch returns a block of memory of length n + 1 from the
// scratch buffer.
func extract_from_scratch(scratch *scratchBuf) []int64 {
	if scratch.l == 0 {
		scratch.l = SCRATCH_SIZE
		scratch.b = make([]int64, (scratch.n+1)*SCRATCH_SIZE)
	}
	a := scratch.b[:scratch.n+1]
	scratch.b = scratch.b[scratch.n+1:]
	scratch.l--
	return a
}

// output_weights outputs the \lambda values to the output writer.
func output_weights(n int, done chan bool, out_c chan []int64) {
	// Select the output writer
	writer := out_w
	if writer == nil {
		writer = os.Stdout
	}
	// Wrap the destination writer in a buffered writer
	b := bufio.NewWriterSize(writer, WRITER_BUF_SIZE)
	// Allocate a block of memory for the \lambda values
	lambda := make([]int64, n+1)
	// Start working through the k values
	for k := range out_c {
		// Have we been asked to exit?
		if k == nil {
			break
		}
		// Calculate h
		h := LcmSlice(k)
		// Calculate the \lambda values
		for i, kval := range k {
			lambda[i] = h / kval
		}
		// Output them
		fmt.Fprintln(writer, lambda)
	}
	// Flush the buffer and indicate that we're finished
	b.Flush()
	done <- true
}

// check_sum_valid checks that s(\kappa) and the fractional sum are valid
func check_sum_valid(k []int64, h int64) bool {
	// Step through all \kappa\in {1,...,h-1} calculating the s value and sum
	max_sum := len(k) - 3
	sum := 1
	for kappa := int64(2); kappa <= h-3; kappa++ {
		// Calculate s(\kappa)
		s := 0
		for _, kval := range k {
			if (kappa % kval) == 0 {
				s++
			}
		}
		// Update the fractional sum for this \kappa
		sum += 1 - s
		// Check that the values are valid
		if sum < 2 || sum > max_sum {
			return false
		}
	}
	// If we're here then it looks good
	return true
}

// calculate_final_lambdas works through the possible values of \lambda_0,
// \lambda_1, and \lambda_2, outputting the possible \lambda-values
func calculate_final_lambdas(k []int64, h int64, lam int64, scratch *scratchBuf, out_c chan []int64) {
	// Establish upper and lower bounds on \lambda_2
	max2 := lam - 2
	if max2 > h/k[3] {
		max2 = h / k[3]
	}
	if m := Floor(h, int64(len(k))); max2 > m {
		max2 = m
	}
	min2 := Ceiling(lam, 3)
	// Now check each possible value of \lambda_2 in turn
	for lambda2 := max2; lambda2 >= min2; lambda2-- {
		// We require \lambda_2 | h
		if (h % lambda2) == 0 {
			// Record the corresponding k-value
			k[2] = h / lambda2
			// Estabilsh upper and lower bounds on \lambda_1
			max1 := lam - lambda2 - 1
			if max1 > lambda2 {
				max1 = lambda2
			}
			min1 := Ceiling(lam-lambda2, 2)
			// Now check each possible value of \lambda_1 in turn
			for lambda1 := max1; lambda1 >= min1; lambda1-- {
				// We require \lambda_1 | h
				if (h % lambda1) == 0 {
					// Record the corresponding k-value
					k[1] = h / lambda1
					// \lambda_0 has to be what's left
					lambda0 := lam - lambda2 - lambda1
					// Check that the value of \lambda_0 is valid
					if (lambda0 <= lambda1) && (lambda0 > 0) &&
						(h%lambda0) == 0 {
						// Record the corresponding k-value
						k[0] = h / lambda0
						// Check that the values are valid and output the weight
						if check_sum_valid(k, h) {
							new_k := extract_from_scratch(scratch)
							copy(new_k, k)
							out_c <- new_k
						}
					}
				}
			}
		}
	}
}

// finish_calculation checks the weight is valid and outputs the \lambda-values
func finish_calculation(k []int64, scratch *scratchBuf, out_c chan []int64) {
	// Calculate the value of h
	ks_so_far := k[3:]
	h := LcmSlice(ks_so_far)
	// Calculate the sum \lambda_3 + ... + \lambda_n
	total := int64(0)
	for _, kval := range ks_so_far {
		total += h / kval
	}
	// And the sum \lambda_0 + \lambda_1 + \lambda_2
	lam := h - total
	if lam < 3 {
		return
	}
	// Work through the possible \lambda_i, i = 0,1,2, outputting any weights
	calculate_final_lambdas(k, h, lam, scratch, out_c)
}

// final_worker is a worker wrapper around finish_calculation for calculating
// the final three \lambda values and outputting the results.
func final_worker(n int, done chan bool, final_c chan []int64, out_c chan []int64) {
	// Allocate the memory for the scratch buffer
	scratch := new_scratch(n)
	// Start working through the jobs
	for k := range final_c {
		// Have we been asked to exit?
		if k == nil {
			break
		}
		// Finish the calculation
		finish_calculation(k, scratch, out_c)
	}
	// Indicate that we're finished
	done <- true
}

// tighten_upper_bound attempts to umprove upon the given upper bound by direct
// calculation of the sum.
func tighten_upper_bound(n int, depth int, k []int64, upper int64) int64 {
	// Work through the \kappa
	sum := int64(1)
	for kappa := int64(2); kappa <= upper-1; kappa++ {
		// Calculate the sum as it currently stands
		sum++
		for i := depth + 1; i <= n; i++ {
			if (kappa % k[i]) == 0 {
				sum--
			}
		}
		// Check that the values are valid
		if (sum > int64(n-1)) && (kappa < upper-2) {
			upper = kappa + 2
		}
		if (sum > int64(n-2)) && (kappa < upper-3) {
			upper = kappa + 3
		}
		if (sum < 2) && (kappa < upper-2) {
			upper = kappa + 2
		}
	}
	// Return the bound we found
	return upper
}

// calculate_k_recursive recursively calculates the possible k-values. It
// expects depth <= n - 2.
func calculate_k_recursive(n int, depth int, k []int64, scratch *scratchBuf, final_c chan []int64) {
	// Are we down to calcuating k_2?
	if depth == 2 {
		new_k := extract_from_scratch(scratch)
		copy(new_k, k)
		final_c <- new_k
		return
	}
	// Otherwise, step through the possible values of k_depth
	// First establish the lower bounds
	lower := int64(n - depth + 3)
	if lower < k[depth+1] {
		lower = k[depth+1]
	}
	// Now the upper bounds
	// We caluclate the bound in stages. First we calculate \sum 1/k_i
	sum := new(big.Rat)
	for i := depth + 1; i <= n; i++ {
		sum.Add(sum, big.NewRat(1, k[i]))
	}
	// If the sum is >= 1, we've already blown it
	unit := big.NewRat(1, 1)
	if sum.Cmp(unit) >= 0 {
		return
	}
	// Calculate 1/(depth + 1) * (1 - sum)
	sum.Sub(unit, sum)
	sum.Mul(big.NewRat(1, int64(depth+1)), sum)
	// Now take the reciprocal
	sum.Inv(sum)
	// The bound is Floor(sum)
	t := sum.Num()
	t.Div(t, sum.Denom())
	upper := t.Int64()
	// Is this range sensible?
	if upper < lower {
		return
	}
	// Can the upper bound can be improved by direct calculation of the sum?
	upper = tighten_upper_bound(n, depth, k, upper)
	// Iterate (and recurse) over this range
	for kval := lower; kval <= upper; kval++ {
		k[depth] = kval
		calculate_k_recursive(n, depth-1, k, scratch, final_c)
	}
}

// calculate_k_worker is a worker wrapper around calculate_k_recursive for
// recursively calcilating the possible k values.
func calculate_k_worker(n int, done chan bool, job_c chan []int64, final_c chan []int64) {
	// Allocate the memory for the scratch buffer
	scratch := new_scratch(n)
	// Start working through the jobs
	for k := range job_c {
		// Have we been asked to exit?
		if k == nil {
			break
		}
		// Start searching
		calculate_k_recursive(n, n-2, k, scratch, final_c)
	}
	// Indicate that we're finished
	done <- true
}

// feed_workers feeds the jobs to the workers.
func feed_workers(n int, job_c chan []int64) {
	// Allocate a scratch buffer and k-values
	scratch := new_scratch(n)
	k := extract_from_scratch(scratch)
	// Loop over the range of values for k_n
	for kn := int64(3); kn <= int64(n-1); kn++ {
		k[n] = kn
		// Calculate the lower bound for k_{n-1}
		lower := kn
		if kn == 3 {
			lower = int64(4)
		}
		// Calculate the upper bound for k_{n-1}
		upper := Floor(int64(n)*kn, kn-1)
		// Is there anything to do?
		if lower <= upper {
			// Can the upper bound can be improved by direct calculation?
			upper = tighten_upper_bound(n, n-1, k, upper)
			if lower <= upper {
				// Feed the jobs to the workers
				for kval := lower; kval <= upper; kval++ {
					k[n-1] = kval
					new_k := extract_from_scratch(scratch)
					copy(new_k, k)
					job_c <- new_k
				}
			}
		}
	}
}

// calculate_k is the main entry function for recursively calculating the
// possible weights. Handles starting up and managing the workers. Assumes
// that n >= 3.
func calculate_k(n int) {
	// We want to limit the number of workers
	num_workers := runtime.NumCPU()
	// Create the local communication channels
	job_c := make(chan []int64, 2*num_workers)
	job_done := make(chan bool, 1)
	final_c := make(chan []int64, 2*num_workers)
	final_done := make(chan bool, 1)
	out_c := make(chan []int64, CHAN_BUF_SIZE)
	out_done := make(chan bool, 1)
	// Start the output go routine
	go output_weights(n, out_done, out_c)
	// Launch the final worker go routines
	for i := 0; i < num_workers; i++ {
		go final_worker(n, final_done, final_c, out_c)
	}
	// Launch the worker go routines
	for i := 0; i < num_workers; i++ {
		go calculate_k_worker(n, job_done, job_c, final_c)
	}
	// Feed the jobs to the workers
	feed_workers(n, job_c)
	// Ask the workers to exit
	for i := 0; i < num_workers; i++ {
		job_c <- nil
	}
	// Wait for the workers to finish
	finished := 0
	for _ = range job_done {
		finished++
		if finished == num_workers {
			break
		}
	}
	// Ask the final workers to exit
	for i := 0; i < num_workers; i++ {
		final_c <- nil
	}
	// Wait for the final workers to finish
	finished = 0
	for _ = range final_done {
		finished++
		if finished == num_workers {
			break
		}
	}
	// Ask the output manager to exit and wait for it to finish
	out_c <- nil
	<-out_done
}

// fatalf outputs the formatted message to Stderr and exits.
func fatalf(format string, a ...interface{}) {
	s := fmt.Sprintf(format, a...)
	fmt.Fprintf(os.Stderr, "%s: %s\n", os.Args[0], strings.TrimSpace(s))
	os.Exit(1)
}

// parse_args parses the command-line arguments.
func parse_args() (args cmdArgs) {
	// Bind the options
	flag.IntVar(&args.n, "n", 4, "The dimension to search in.")
	flag.StringVar(&args.out, "out", "", "Output weights to the given file.")
	// Parse the arguments
	flag.Parse()
	// Validate the arguments
	if args.n <= 3 {
		fatalf("The dimension must be at least 4.")
	}
	return
}

/////////////////////////////////////////////////////////////////////////
// main
/////////////////////////////////////////////////////////////////////////

// main is the program entry/exit point.
func main() {
	// Use all the machine's cores
	runtime.GOMAXPROCS(runtime.NumCPU())
	// Parse the arguments
	args := parse_args()
	// Create the output file
	if args.out != "" {
		var err error
		if out_w, err = os.Create(args.out); err != nil {
			fatalf("Unable to create output file [%s].", err)
		}
	}
	// Start the search
	calculate_k(args.n)
}
